/*
 * finddm.c - Find a drive mode change algorithm by brute force
 *
 * Written 2006 by Werner Almesberger
 * Copyright 2006 Werner Almesberger
 */


#include <stdio.h>


#define FILTER_ANY (1)
#define FILTER_DR0 ((a & 1) == 0 && (b & 1) == 0)
#define FILTER_DR1 ((a & 1) == 1 && (b & 1) == 1)
#define FILTER_DRx ((a^b) & 1)
#define FILTER_DR0_1 ((a & 1) == 0 && (b & 1) == 1)
#define FILTER_DR1_0 ((a & 1) == 1 && (b & 1) == 0)

#ifndef FILTER
#define FILTER FILTER_ANY
#endif

#ifndef COND
#define COND monotonous
#endif


#define MAX_DISP 4


enum value {
    VALUE_0,
    VALUE_0R,
    VALUE_Z,
    VALUE_1R,
    VALUE_1,
};


static const enum value value[] = {
    VALUE_0R, VALUE_1,
    VALUE_0,  VALUE_1,
    VALUE_Z,  VALUE_Z,
    VALUE_0,  VALUE_1R,
    VALUE_Z,  VALUE_1,
    VALUE_0,  VALUE_1,
    VALUE_Z,  VALUE_Z,
    VALUE_0,  VALUE_Z,
};


/*
 * It's either MOV, SET+MOV or CLR+MOV, never SET+CLR or CLR+SET
 */

static int op_mask[] = {
    0x01, 0x02, 0x04, 0x08,	/* SET */
    0x01, 0x02, 0x04, 0x08,	/* CLR */
    0x11, 0x22, 0x44, 0x88,	/* MOV */
};

#define SET(n)	#n " |= new_" #n ";"
#define CLR(n)	#n " &= new_" #n ";"
#define MOV(n)	#n " = new_" #n ";"

static const char *op_name[] = {
    SET(dr), SET(dm0), SET(dm1), SET(dm2),
    CLR(dr), CLR(dm0), CLR(dm1), CLR(dm2),
    MOV(dr), MOV(dm0), MOV(dm1), MOV(dm2),
};


static int available = 0xfff;
static int best = 999;
static int ops[8];


static void report(int level)
{
    int i;

    printf("----- (cost %d)\n",best);
    for (i = 0; i != level+1; i++)
	printf("%s\n",op_name[ops[i]]);
    fflush(stdout);
}


static int dir = 0;


static int __attribute__((unused)) monotonous(int va,int vb,int v,int n)
{
    if (n == v)
	return 1;
    if (n < v) {
	if (dir == 1)
	    return 0;
	dir = -1;
    }
    else {
	if (dir == -1)
	    return 0;
	dir = 1;
    }
    return 1;
}


static int __attribute__((unused)) nonextremal(int va,int vb,int v,int n)
{
    if (n != VALUE_0 && n != VALUE_1)
	return 1;
    return n == va || n == vb;
}


static int try(int a,int b,int level)
{
    int va = value[a];
    int vb = value[b];
    int v,i,m;

    dir = 0;
    v = va;
    for (i = 0; i != level; i++) {
	int op,n;

	op = ops[i];
	if (op < 4) /* SET */
	    a |= b & (1 << op);
	else if (ops[i] < 8) /* CLR */
	    a &= b | ~(1 << (op-4));
	else { /* MOV */
	    m = 1 << (op-8);
	    a = (a & ~m) | (b & m);
	}
	n = value[a];
	if (!COND(va,vb,v,n))
	    return -1;
	v = n;
    }
    return a == b;
}


static int xio(int op)
{
    op &= 3;
    return op == 1 || op == 2; /* DM0 or DM1 */
}


static int cycles(op)
{
    if (op < 8)
	return 9;	/* OR reg[expr],expr */
    else
	return 8;	/* MOV reg[expr],expr */
}


static void do_level(int level,int cost)
{
    int i;

    if (level == 8)
	return;
    for (i = 0; i != 12; i++) {
	if (level < MAX_DISP)
	    fprintf(stderr,"%x",i);
	if (op_mask[i] & available) {
	    int found = 1;
	    int this_cost,xio_cost,old_xio,new_xio;
	    int a,b;

	    this_cost = cycles(i);
	    old_xio = level ? xio(ops[level-1]) : 0;
	    new_xio = xio(i);
	    if (old_xio != new_xio)
		this_cost += 4; /* OR F,expr */
	    xio_cost = new_xio ? 4 : 0; /* AND F,~expr */
	    if (cost+this_cost+xio_cost < best) {
		ops[level] = i;
		available &= ~op_mask[i];
		for (a = 0; a < 16; a++)
		    for (b = 0; b < 16; b++)
			if (FILTER)
			    switch (try(a,b,level+1)) {
				case -1:
				    found = 0;
				    goto fail;
				case 0:
				    found = 0;
				    break;
				case 1:
				    break;
			    }
		if (!found)
		    do_level(level+1,cost+this_cost);
fail:
		available |= op_mask[i];
		if (found) {
		    best = cost+this_cost+xio_cost;
		    report(level);
		    return;
		}
	    }
	}
	if (level < MAX_DISP)
	    fprintf(stderr,"\b");
    }
}


int main(void)
{
    do_level(0,0);
    return 0;
}
